@@ -1,46 +1,22 @@
-# coding: utf-8
-# /*##########################################################################
-#
-# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
-#
-# Permission is hereby granted, free of charge, to any person obtaining a copy
-# of this software and associated documentation files (the "Software"), to deal
-# in the Software without restriction, including without limitation the rights
-# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-# copies of the Software, and to permit persons to whom the Software is
-# furnished to do so, subject to the following conditions:
-#
-# The above copyright notice and this permission notice shall be included in
-# all copies or substantial portions of the Software.
-#
-# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-# THE SOFTWARE.
-#
-# ###########################################################################*/
+from __future__ import annotations
 
-
-__authors__ = ["H. Payno"]
-__license__ = "MIT"
-__date__ = "03/11/2020"
-
-
-from typing import Union
-from silx.gui import qt
+import fnmatch
 from collections import OrderedDict
+from typing import Union
+import datetime
+
+from silx.gui import qt
+
+from processview.core.dataset import DatasetIdentifier
 from processview.core.manager import DatasetState
 from processview.core.manager import ProcessManager as _ProcessManager
+from processview.core.sorting import SortType, tooltips
 from processview.core.superviseprocess import SuperviseProcess
-from processview.core.dataset import DatasetIdentifier
+from processview.gui.DropDownWidget import DropDownWidget
 from processview.gui import icons as icons
 from processview.gui.messagebox import MessageBox
 from processview.utils import docstring
-import fnmatch
-
+from processview.gui.utils.qitem_model_resetter import qitem_model_resetter
 
 _DATASET_STATE_BACKGROUND = {
     DatasetState.ON_GOING: qt.QColor("#839684"),  # light blue
@@ -49,6 +25,7 @@ _DATASET_STATE_BACKGROUND = {
     DatasetState.PENDING: qt.QColor("#609ab3"),  # blue gray
     DatasetState.SKIPPED: qt.QColor("#f08e0e"),  # light orange
     DatasetState.WAIT_USER_VALIDATION: qt.QColor("#cb34c1"),  # pink
+    DatasetState.CANCELLED: qt.QColor("#a4a8a2"),  # light black
 }
 
 
@@ -66,82 +43,110 @@ class ProcessManagerWindow(qt.QMainWindow):
 
 
 class ObservationTable(qt.QTableView):
+    """
+    Redefinition of QTableView for datasets and processes
+    """
+
     def __init__(self, parent):
         qt.QTableView.__init__(self, parent)
         self.verticalHeader().setSectionsClickable(True)
 
+        # QMenu for the dataset name
+        self.dataset_menu = qt.QMenu()
+        self._copyAction = qt.QAction("copy")
+        self.dataset_menu.addAction(self._copyAction)
+        self._removeAction = qt.QAction("remove")
+        self.dataset_menu.addAction(self._removeAction)
+
         # QMenu for cell from on dataset and one process
         self.menu_dataset_vs_process = qt.QMenu()
         self._reprocessAction = qt.QAction("reprocess")
         self.menu_dataset_vs_process.addAction(self._reprocessAction)
+        self._cancelAction = qt.QAction("cancel")
+        self.menu_dataset_vs_process.addAction(self._cancelAction)
         self._infoAction = qt.QAction("info")
         self.menu_dataset_vs_process.addAction(self._infoAction)
+        self.menu_dataset_vs_process.addAction(self._removeAction)
 
         self._target = (None, None)
         # register target of the last menu (process, DatasetIdentifier)
 
-        self.copy_menu = qt.QMenu()
-        self._copyAction = qt.QAction("copy")
-        self.copy_menu.addAction(self._copyAction)
-
         # connect signal / slot
         self._copyAction.triggered.connect(self._requestDatasetIdCopy)
+        self._removeAction.triggered.connect(self._requestRemoveDataset)
         self._reprocessAction.triggered.connect(self._requestReprocessing)
         self._infoAction.triggered.connect(self._requestInfo)
-
-        self.verticalHeader().setContextMenuPolicy(qt.Qt.CustomContextMenu)
-        self.verticalHeader().customContextMenuRequested.connect(
-            self._activeVerticalHeaderMenu
-        )
+        self._cancelAction.triggered.connect(self._requestCancelProcessing)
 
     def _processAt(self, x_pos):
         column = self.columnAt(x_pos)
-        if column >= 0:
+        if column >= 1:
             processes = self.model()._processes
-            if column < len(processes):
-                return processes[list(processes.keys())[column]]
+            process_idx = column - 1
+            if process_idx < len(processes):
+                return processes[list(processes.keys())[process_idx]]
 
     def _datasetAt(self, y_pos):
         row = self.rowAt(y_pos)
         if row >= 0:
-            datasets = self.model()._datasets
+            datasets = self.model()._sorted_datasets
             if row < len(datasets):
                 return datasets[list(datasets.keys())[row]]
 
-    def _activeVerticalHeaderMenu(self, pos):
-        process = self._processAt(pos.x())
-        dataset = self._datasetAt(pos.y())
-        if dataset is not None:
-            self._target = (None, dataset)
-            self.copy_menu.exec_(self.mapToGlobal(pos))
-
     def contextMenuEvent(self, event):
-        process = self._processAt(event.pos().x())
+        row = self.columnAt(event.pos().x())
         dataset = self._datasetAt(event.pos().y())
-        if (
-            process is not None
-            and dataset is not None
-            and _ProcessManager().met(process=process, dataset=dataset)
-        ):
-            self._target = (process, dataset)
-            self.menu_dataset_vs_process.exec_(event.globalPos())
+        if row == 0:
+            # handle column column
+            self._target = (None, dataset)
+            self.dataset_menu.exec_(event.globalPos())
         else:
-            self._target = (None, None)
+            # handle processes column
+            process = self._processAt(event.pos().x())
+            if (
+                process is not None
+                and dataset is not None
+                and _ProcessManager().met(process=process, dataset=dataset)
+            ):
+                self._target = (process, dataset)
+                self.menu_dataset_vs_process.exec_(event.globalPos())
+            else:
+                self._target = (None, None)
         super().contextMenuEvent(event)
 
     def _requestReprocessing(self, *args, **kwargs):
         process, dataset = self._target
+
         if process is not None and dataset is not None:
             assert isinstance(process, SuperviseProcess)
             assert isinstance(dataset, DatasetIdentifier)
             process.reprocess(dataset.recreate_dataset())
 
+    def _requestCancelProcessing(self, *args, **kwargs):
+        process, dataset = self._target
+        if process is not None and dataset is not None:
+            assert isinstance(process, SuperviseProcess)
+            assert isinstance(dataset, DatasetIdentifier)
+            process.cancel(dataset.recreate_dataset())
+
     def _requestDatasetIdCopy(self, *args, **kwargs):
         _, dataset = self._target
         if dataset is not None:
             clipboard = qt.QGuiApplication.clipboard()
             clipboard.setText(dataset.to_str())
 
+    def _requestRemoveDataset(self, *args, **kwargs):
+        def get_dataset_at(row: int):
+            datasets = self.model()._sorted_datasets
+            return datasets.get(
+                list(datasets.keys())[row],
+                None,
+            )
+
+        datasets = [get_dataset_at(index.row()) for index in self.selectedIndexes()]
+        [ProcessManager().remove_dataset(dataset) for dataset in datasets]
+        self.model().remove_datasets(datasets)
+
     def _requestInfo(self, *args, **kwargs):
         process, dataset = self._target
         if process is not None and dataset is not None:
@@ -158,21 +163,38 @@ class ObservationTable(qt.QTableView):
             msg.setWindowModality(qt.Qt.NonModal)
             msg.show()
 
+    def sizeHintForColumn(self, column):
+        if column == 0:
+            return 350
+        else:
+            return super().sizeHintForColumn(column)
+
 
 class ProcessManagerWidget(qt.QWidget):
+    """
+    Main widget to display dataset vs process metadata
+    """
+
     def __init__(self, parent):
         qt.QWidget.__init__(self, parent)
         self.setLayout(qt.QVBoxLayout())
 
         self._manager = ProcessManager()
 
-        self.filterWidget = _ToggleableFilterWidget(parent=self)
-        self.layout().setContentsMargins(4, 4, 4, 4)
-        self.layout().setSpacing(4)
-        self.filterWidget.layout().setSpacing(4)
-        self.layout().addWidget(self.filterWidget)
+        self._optionsWidget = OptionsWidget(parent=self)
+        self._dropDownOptionsWidget = DropDownWidget(parent=self)
+        self._dropDownOptionsWidget.setWidget(self._optionsWidget)
+
+        self._dropDownOptionsWidget.setSizePolicy(
+            qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum
+        )
+        self.setContentsMargins(0, 0, 0, 0)
+        self.layout().setContentsMargins(0, 0, 0, 0)
+        self.layout().setSpacing(2)
+        self.layout().addWidget(self._dropDownOptionsWidget)
 
         self.observationTable = ObservationTable(self)
+
         self.layout().addWidget(self.observationTable)
         self.observationTable.setModel(
             _DatasetProcessModel(parent=self.observationTable, header=tuple())
@@ -184,13 +206,17 @@ class ProcessManagerWidget(qt.QWidget):
         # connect signal / slot
         self._manager.sigUpdated.connect(self._updateDatasetStates)
         self._manager.sigNewProcessRegistered.connect(self._updateProcesses)
-        self.filterWidget.filterWidget.sigDatasetPatternEditingFinished.connect(
+        self._optionsWidget.filterWidget.sigDatasetPatternEditingFinished.connect(
             self._filterUpdated
         )
-        self.filterWidget.filterWidget.sigProcessPatternEditingFinished.connect(
+        self._optionsWidget.filterWidget.sigProcessPatternEditingFinished.connect(
             self._filterUpdated
         )
         self._manager.sigProcessUnregistered.connect(self._updateProcesses)
+        self._optionsWidget._orderingWidget.sigSortTypeChanged.connect(
+            self.observationTable.model()._setSorting
+        )
+        self._manager.sigProcessRenamed.connect(self._renameProcess)
         # update to fit existing processes / datasets
         self._updateProcesses()
         self._updateDatasetStates()
@@ -201,13 +227,18 @@ class ProcessManagerWidget(qt.QWidget):
     def _updateProcesses(self):
         self.observationTable.model().setProcesses(self._manager.get_processes())
 
+    def _renameProcess(self, process: SuperviseProcess):
+        self.observationTable.model().headerDataChanged.emit(
+            qt.Qt.Horizontal, 0, len(self._manager.get_processes())
+        )
+
     def _filterUpdated(self):
         self.observationTable.model().process_patterns = (
-            self.filterWidget.getProcessPatterns()
+            self._dropDownOptionsWidget.getProcessPatterns()
         )
         self._updateProcesses()
         self.observationTable.model().dataset_patterns = (
-            self.filterWidget.getDatasetPatterns()
+            self._dropDownOptionsWidget.getDatasetPatterns()
         )
         self._updateDatasetStates()
 
@@ -217,11 +248,14 @@ class _DatasetProcessModel(qt.QAbstractTableModel):
         qt.QAbstractTableModel.__init__(self, parent, *args)
         self.header = header
         self._processes = OrderedDict()
-        self._datasets = OrderedDict()
+        # processes with order as key and Process as value
+        self._sorted_datasets = OrderedDict()
+        self._unsorted_datasets = OrderedDict()
         self._processPatterns = tuple()
         # is there some process name pattern to follow ?
         self._datasetPatterns = tuple()
         # is there some dataset id pattern to follow ?
+        self._sort_type = SortType.FIRST_APPEARANCE
 
     def _match_dataset_patterns(self, dataset):
         if len(self._datasetPatterns) == 0:
@@ -257,57 +291,92 @@ class _DatasetProcessModel(qt.QAbstractTableModel):
     def dataset_patterns(self, patterns):
         self._datasetPatterns = patterns
 
-    def add(self, dataset, status):
-        self._processes[dataset] = status
-        self.endResetModel()
+    def remove_datasets(self, datasets):
+        remaining_datasets = list(self._unsorted_datasets)
+        for dataset in datasets:
+            if not isinstance(dataset, DatasetIdentifier):
+                raise ValueError(
+                    f"dataset is expected to be a dataset identifier. Get {type(dataset)} instead"
+                )
+            try:
+                remaining_datasets.remove(dataset)
+            except Exception:
+                pass
 
-    def remove(self, dataset):
-        if dataset in self._processes:
-            del self._processes[dataset]
-        self.endResetModel()
-
-    def update_dataset(self, dataset, status):
-        self._processes[dataset] = status
-        self.endResetModel()
+        with qitem_model_resetter(self):
+            self.setDatasets(remaining_datasets)
 
     def clear(self):
-        self._processes = OrderedDict()
-        self.endResetModel()
+        with qitem_model_resetter(self):
+            self._processes = OrderedDict()
 
     def rowCount(self, parent=None):
-        return len(self._datasets)
+        return len(self._unsorted_datasets)
 
     def columnCount(self, parent=None):
-        return len(self._processes)
+        return len(self._processes) + 1
 
     def setProcesses(self, processes):
         assert isinstance(processes, (tuple, list))
-        self._processes = {}
-        processes = list(filter(self._match_process_patterns, processes))
-        for i_process, process in enumerate(processes):
-            self._processes[i_process] = process
-        self.endResetModel()
+        with qitem_model_resetter(self):
+            self._processes = {}
+            processes = list(filter(self._match_process_patterns, processes))
+            for i_process, process in enumerate(processes):
+                self._processes[i_process] = process
+
+    @staticmethod
+    def sort_datasets(datasets: list, sort_type):
+        if sort_type is SortType.FIRST_APPEARANCE:
+            pass
+        elif sort_type is SortType.LAST_APPEARANCE:
+            datasets.reverse()
+        elif sort_type in (SortType.ALPHABETICAL, SortType.REVERSE_ALPHABETICAL):
+            datasets = sorted(
+                datasets,
+                key=lambda x: x.name() or "",
+                reverse=sort_type is SortType.REVERSE_ALPHABETICAL,
+            )
+        elif sort_type in (SortType.CREATION_TIME, SortType.REVERSE_CREATION_TIME):
+            datasets = sorted(
+                datasets,
+                key=lambda x: x.creation_time() or datetime.datetime.fromtimestamp(0),
+                reverse=sort_type is SortType.REVERSE_CREATION_TIME,
+            )
+        elif sort_type in (
+            SortType.MODIFICATION_TIME,
+            SortType.REVERSE_MODIFICATION_TIME,
+        ):
+            datasets = sorted(
+                datasets,
+                key=lambda x: x.modification_time()
+                or datetime.datetime.fromtimestamp(0),
+                reverse=sort_type is SortType.REVERSE_MODIFICATION_TIME,
+            )
+        else:
+            raise ValueError("sort type not handled")
+        return datasets
 
     def setDatasets(self, datasets):
         assert isinstance(datasets, (tuple, list))
-        self._datasets = {}
-        datasets = list(filter(self._match_dataset_patterns, datasets))
-        for i_dataset, dataset in enumerate(datasets):
-            self._datasets[i_dataset] = dataset
-        self.endResetModel()
+        with qitem_model_resetter(self):
+            self._unsorted_datasets = datasets
+            # filter datasets
+            datasets = list(filter(self._match_dataset_patterns, datasets))
+            # sort datasets
+            datasets = self.sort_datasets(datasets=datasets, sort_type=self._sort_type)
+
+            self._sorted_datasets.clear()
+            for i_dataset, dataset in enumerate(datasets):
+                self._sorted_datasets[i_dataset] = dataset
 
     def headerData(self, col, orientation, role):
         if orientation == qt.Qt.Horizontal and role == qt.Qt.DisplayRole:
-            if col < len(self._processes):
-                return self._processes[col].name
-        elif orientation == qt.Qt.Vertical:
-            if col < len(self._datasets):
-                if role == qt.Qt.DisplayRole:
-                    return self._datasets[col].short_description() or str(
-                        self._datasets[col]
-                    )
-                elif role == qt.Qt.ToolTipRole:
-                    return self._datasets[col].long_description()
+            if col == 0:
+                return
+            else:
+                process_idx = col - 1
+            if process_idx < len(self._processes):
+                return self._processes[process_idx].name
         return None
 
     def data(self, index, role):
@@ -317,8 +386,21 @@ class _DatasetProcessModel(qt.QAbstractTableModel):
         if role not in (qt.Qt.DisplayRole, qt.Qt.ToolTipRole, qt.Qt.BackgroundRole):
             return None
 
-        dataset_short_name = self._datasets[index.row()]
-        process_name = self._processes[index.column()]
+        if index.column() == 0:
+            # if dataset name
+            if role == qt.Qt.DisplayRole:
+                return self._sorted_datasets[index.row()].short_description() or str(
+                    self._sorted_datasets[index.row()]
+                )
+            elif role == qt.Qt.ToolTipRole:
+                return self._sorted_datasets[index.row()].long_description()
+            elif role == qt.Qt.BackgroundRole:
+                return qt.QColor("#dedede")
+            else:
+                return
+
+        dataset_short_name = self._sorted_datasets[index.row()]
+        process_name = self._processes[index.column() - 1]
         dataset_process_state = ProcessManager().get_dataset_state(
             dataset=dataset_short_name, process=process_name
         )
@@ -342,46 +424,43 @@ class _DatasetProcessModel(qt.QAbstractTableModel):
             else:
                 return dataset_process_details
 
+    def _setSorting(self, sort_type: SortType):
+        sort_type = SortType.from_value(sort_type)
+        changed = self._sort_type is not sort_type
+        if changed:
+            self._sort_type = sort_type
+            self.setDatasets(datasets=self._unsorted_datasets)
+            self.layoutChanged.emit()
 
-class _ToggleableFilterWidget(qt.QWidget):
 
-    _BUTTON_ICON = qt.QStyle.SP_ToolBarVerticalExtensionButton  # noqa
+class OptionsWidget(qt.QWidget):
 
     def __init__(self, parent):
-        qt.QWidget.__init__(self, parent)
+        super().__init__(parent)
         self.setLayout(qt.QGridLayout())
-        self._toggleButton = qt.QPushButton("", self)
-        self.layout().addWidget(self._toggleButton, 0, 0, 1, 1)
+
+        # dataset ordering
+        self._orderingWidget = _OrderingWidget(parent=self)
+        self.layout().setContentsMargins(2, 2, 2, 2)
+        self.layout().setSpacing(0)
+        self.layout().addWidget(self._orderingWidget, 2, 1, 3, 3)
+
+        # filtering
         self._filterWidget = _FilterWidget(parent=self)
-        self.layout().addWidget(self._filterWidget, 0, 1, 4, 4)
-
-        self._toggleButton.setSizePolicy(qt.QSizePolicy.Fixed, qt.QSizePolicy.Fixed)
-        self._setButtonIcon(show=True)
-
-        # connect signal / slot
-        self._toggleButton.clicked.connect(self._toggleFilterWidget)
-
-        # expose API
-        self.getProcessPatterns = self._filterWidget.getProcessPatterns
-        self.getDatasetPatterns = self._filterWidget.getDatasetPatterns
+        self.layout().setContentsMargins(2, 2, 2, 2)
+        self.layout().setSpacing(0)
+        self.layout().addWidget(self._filterWidget, 4, 1, 3, 3)
 
     @property
     def filterWidget(self):
         return self._filterWidget
 
-    def _setButtonIcon(self, show):
-        style = qt.QApplication.instance().style()
-        # return a QIcon
-        icon = style.standardIcon(self._BUTTON_ICON)
-        if show is True:
-            pixmap = icon.pixmap(32, 32).transformed(qt.QTransform().scale(1, -1))
-            icon = qt.QIcon(pixmap)
-        self._toggleButton.setIcon(icon)
+    # expose API
+    def getProcessPatterns(self):
+        return self._filterWidget.getProcessPatterns()
 
-    def _toggleFilterWidget(self):
-        visible = not self._filterWidget.isVisible()
-        self._setButtonIcon(show=visible)
-        self._filterWidget.setVisible(visible)
+    def getDatasetPatterns(self):
+        return self._filterWidget.getDatasetPatterns()
 
 
 class _FilterWidget(qt.QWidget):
@@ -395,9 +474,11 @@ class _FilterWidget(qt.QWidget):
     sigDatasetPatternEditingFinished = qt.Signal()
     """signal emit when the dataset pattern editing finished"""
 
-    def __init__(self, parent=None, name="filter", font_size=12, icon_size=20):
+    def __init__(self, parent=None, name=" filter", font_size=12, icon_size=20):
         qt.QWidget.__init__(self, parent)
         self.setLayout(qt.QGridLayout())
+        self.layout().setSpacing(2)
+        self.layout().setContentsMargins(0, 0, 0, 0)
 
         font = self.font()
         font.setPixelSize(font_size)
@@ -480,6 +561,40 @@ class _FilterWidget(qt.QWidget):
         self.sigProcessPatternEditingFinished.emit()
 
 
+class _OrderingWidget(qt.QWidget):
+    """Widget to let the user define the ordering of datasets"""
+
+    sigSortTypeChanged = qt.Signal(str)
+    """emit when sort type changed. Paramater is the name of the new sorting type"""
+
+    def __init__(self, parent=None, *args, **kwargs) -> None:
+        super().__init__(parent, *args, **kwargs)
+        self.setLayout(qt.QFormLayout())
+        self._sortingTypeCB = qt.QComboBox(self)
+        for sort_type in SortType.members():
+            self._sortingTypeCB.addItem(sort_type.value)
+            self._sortingTypeCB.setItemData(
+                self._sortingTypeCB.findText(sort_type.value),
+                tooltips[sort_type],
+                qt.Qt.ToolTipRole,
+            )
+        self.layout().addRow("sorting", self._sortingTypeCB)
+
+        # set up
+        self._sortingTypeCB.setCurrentIndex(
+            self._sortingTypeCB.findText(SortType.FIRST_APPEARANCE.value)
+        )
+
+        # connect signal / slot
+        self._sortingTypeCB.currentIndexChanged.connect(self._changed)
+
+    def _changed(self, *args, **kwargs):
+        self.sigSortTypeChanged.emit(self.getCurrentSort().value)
+
+    def getCurrentSort(self) -> SortType:
+        return SortType.from_value(self._sortingTypeCB.currentText())
+
+
 class ProcessManager(qt.QObject):
     sigUpdated = qt.Signal()
     """Signal emitted when the state of some process / dataset is updated
@@ -491,6 +606,9 @@ class ProcessManager(qt.QObject):
     sigProcessUnregistered = qt.Signal()
     """Signal emitted when a process is unregistered"""
 
+    sigProcessRenamed = qt.Signal(SuperviseProcess)
+    """Emit when the process is renamed"""
+
     def __init__(self):
         qt.QObject.__init__(self)
         self.manager = _ProcessManager()
@@ -500,6 +618,7 @@ class ProcessManager(qt.QObject):
         self.manager.add_update_callback(self.updated)
         self.manager.add_new_process_callback(self.processAdded)
         self.manager.add_process_removed_callback(self.processRemoved)
+        self.manager.add_process_name_changed_callback(self.processRenamed)
 
     def updated(self):
         self.sigUpdated.emit()
@@ -510,8 +629,8 @@ class ProcessManager(qt.QObject):
     def processRemoved(self):
         self.sigProcessUnregistered.emit()
 
-    def processNameChanged(self, process):
-        self.sigProcessNameChanged.emit(process)
+    def processRenamed(self, process):
+        self.sigProcessRenamed.emit(process)
 
     def destroyed(self, object_):
         self.manager.remove_update_callback(self.updated)
@@ -547,3 +666,6 @@ class ProcessManager(qt.QObject):
     @docstring(_ProcessManager)
     def get_datasets(self):
         return self.manager.get_datasets()
+
+    def remove_dataset(self, *args, **kwargs):
+        return self.manager.remove_dataset(*args, **kwargs)
